class FeatureNotFoundError(Exception):
    """
    Exception raised when a requested feature is not found.

    Attributes:
        feature_name (str): The name of the feature that was not found.
    """
    def __init__(self, feature_name):
        self.feature_name = feature_name
        super().__init__(f"Feature name: {feature_name} is Not Found")


class InvalidMoleculeTypeError(Exception):
    """
    Exception raised when attempting to calculate a feature in the molecule 
    where it is not possible due to the impossibility of calculating X1, X2, R1 or R2.

    Attributes:
        X1 (int): Atom id in cycle, that connects NH2 or COOH to the molecule.
        X2 (int): Atom id in cycle, that connects fluorine functional group to the molecule.
        R1 (int): NH2 or COOH functional group center atom id.
        R2 (int): fluorine functional group center atom id.
        feature_name (str): The name of the feature.
    """
    def __init__(self, X1, X2, R1, R2, feature_name):
        self.X1 = X1
        self.X2 = X2
        self.R1 = R1
        self.R2 = R2
        self.feature_name = feature_name
        super().__init__(f"Cannot calculate: {feature_name} due to invalid values:\n\
                         X1: {X1}\n\
                         X2: {X2}\n\
                         R1: {R1}\n\
                         R2: {R2}")



import math
import numpy as np
from collections import deque

from rdkit import Chem
from rdkit.Chem import rdchem, rdMolTransforms, rdForceFieldHelpers, rdPartialCharges, rdFreeSASA
from rdkit.Chem import AllChem, Descriptors
from rdkit.Geometry import Point3D
from rdkit.Chem.rdchem import RWMol

from fluoriclogppka.ml_part.constants import Identificator, Target
from fluoriclogppka.ml_part.constants import ALL_SUBMOLS, FUNCTIONAL_GROUP_TO_SMILES
import fluoriclogppka.ml_part.services.utils as utils
import fluoriclogppka.ml_part.services.utils_pKa as utils_pKa
import fluoriclogppka.ml_part.services.utils_logP as utils_logP

class Molecule3DFeaturesService:
    def __init__(self, 
                 smiles: str,
                 target_value: Target) -> None:

        self.target_value = target_value
        self.mol_2d = Chem.MolFromSmiles(smiles)
        self.smiles = smiles
        self.mol = Molecule3DFeaturesService.prepare_molecule(smiles)
        self.min_energy_conf_index, self.min_energy, self.mol = Molecule3DFeaturesService.find_conf_with_min_energy(self.mol)
        self.mol_optimized = self.mol

        self.f_group = self.calculate_fluoric_group()
        self.f_freedom = self.calculate_f_group_freedom()

        self.identificator = self.calculate_identificator()
        self.f_to_fg = self.calculate_linear_path_f_to_fg()
        
        self.dipole_moment = self.calculate_dipole_moment()
        self.mol_volume = self.calculate_volume()
        self.sasa = self.calculate_sasa()
        self.molecular_weight = self.calculate_molecular_weight()

        self.tpsa_with_fluor = self.calculate_TPSA_with_fluor()

        if self.f_group is not None and self.identificator is not None:
            self.X1, self.X2, self.R1, self.R2 = self.find_X1X2R1R2()
            self.dihedral_angle_value = self.calculate_dihedral_angle()
            
            self.distance_between_atoms_in_cycle = self.calculate_distance_between_atoms_in_cycle()
            self.distance_between_atoms_in_f_group_centers = self.calculate_distance_between_atoms_in_f_group_centers()

            self.flat_angle_between_atoms_in_cycle_1, self.flat_angle_between_atoms_in_cycle_2 = self.calculate_flat_angle_between_atoms_in_cycle()
            self.flat_angle_between_atoms_in_f_group_center_1, self.flat_angle_between_atoms_in_f_group_center_2 = self.calculate_flat_angle_between_atoms_in_f_group_center()

            self.cis_trans = self.calculate_cis_trans()

        self.features_3d_dict = {
            "identificator": self.identificator,
            "dipole_moment": self.dipole_moment,
            "mol_volume": self.mol_volume,
            "mol_weight": self.molecular_weight,
            "f_to_fg": self.f_to_fg,
            "sasa": self.sasa,

            "f_freedom": self.f_freedom,

            "cis/trans": self.cis_trans,
            "dihedral_angle": self.dihedral_angle_value,

            "distance_between_atoms_in_cycle_and_f_group": self.distance_between_atoms_in_cycle,
            "distance_between_atoms_in_f_group_centers": self.distance_between_atoms_in_f_group_centers,

            "angle_X1X2R2": self.flat_angle_between_atoms_in_cycle_1,
            "angle_X2X1R1": self.flat_angle_between_atoms_in_cycle_2,

            "angle_R2X2R1": self.flat_angle_between_atoms_in_f_group_center_1,
            "angle_R1X1R2": self.flat_angle_between_atoms_in_f_group_center_2,

            "cis/trans": self.cis_trans,

            "tpsa+f": self.tpsa_with_fluor
        }

    @staticmethod
    def prepare_molecule(smiles):
        """
        Create rdkit 3d molecule from SMILES.
        Generate charges and molecule's conformers.
        
        Args:
            smiles (str): String representation of a molecule. 
            
        Returns:
            mol: Rdkit sanitized molecule with generated charges and multiple conformers.
        """
        mol = Chem.MolFromSmiles(smiles)
        mol = Chem.AddHs(mol)
        rdForceFieldHelpers.MMFFSanitizeMolecule(mol)
        
        num_rotatable_bonds = Descriptors.NumRotatableBonds(mol)
        amount_of_confs = pow(3, num_rotatable_bonds + 3)
        AllChem.EmbedMultipleConfs(mol, numConfs=amount_of_confs, randomSeed=3407)

        rdPartialCharges.ComputeGasteigerCharges(mol)

        return mol

    @staticmethod
    def find_conf_with_min_energy(mol):
        """
        Optimizes all molecules conformers and finds the one with the lowest energy.
        
        Args:
            mol: 3D sanitized molecule with multiple conformers. 

        Returns:
            min_energy_conf_index (int): Conformer index with lowest energy.
            min_energy (float): The lowest conformer energy of the optimized molecule.
            mol: Rdkit optimized molecule.
        """
        optimization_result = rdForceFieldHelpers.MMFFOptimizeMoleculeConfs(mol)
        
        min_energy, min_energy_conf_index = pow(10,5), None
        for index, (status, energy) in enumerate(optimization_result):
            if energy < min_energy and status == 0:
                min_energy_conf_index = index
                min_energy = min(min_energy, energy)

        return min_energy_conf_index, min_energy, mol

    @staticmethod
    def _amount_of_hydrogen_in_neighbors(mol, atom_idx):
        """
        Calculate the number of hydrogen atoms in the neighbors of a given atom.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            atom_idx (int): Index of the target atom.

        Returns:
            int: Number of hydrogen atoms in the neighboring atoms.
        """
        atom = mol.GetAtomWithIdx(atom_idx)

        neighbors = atom.GetNeighbors()
        
        amount_of_hydrogen = 0
        
        for neighbor in neighbors:
            if neighbor.GetSymbol() == 'H':
                amount_of_hydrogen += 1

        return amount_of_hydrogen

    @staticmethod
    def _check_gem_CF2(mol):
        """
        Check if the molecule contains a geminal difluoromethylene group (gem-CF2) 
            based on amount of hydrogen near "C" atom. If more that 0, it's not gem-CF2.

        Args:
            mol (rdchem.Mol): RDKit molecule.

        Returns:
            bool: True if gem-CF2 is present, False otherwise.
        """
        f_group_submol = Chem.MolFromSmiles("C(F)(F)")
        f_group_matches = mol.GetSubstructMatches(f_group_submol)

        if len(f_group_matches) == 0:
            return False
        
        c_atom_index = f_group_matches[0][0]

        amount_of_hydrogen = Molecule3DFeaturesService._amount_of_hydrogen_in_neighbors(mol, c_atom_index)

        if amount_of_hydrogen > 0:
            return False
        
        return True
    
    @staticmethod
    def _check_CH2F(mol):
        """
        Check if the molecule contains a CH2F group based on amount of hydrogen 
            near "C" atom. If "C" near fluorine hasn't 2 hydrogen atoms it's not CH2F.

        Args:
            mol (rdchem.Mol): RDKit molecule.

        Returns:
            bool: True if CH2F is present, False otherwise.
        """
        f_group_submol = Chem.MolFromSmiles("CCF")
        f_group_matches = mol.GetSubstructMatches(f_group_submol)

        if len(f_group_matches) == 0:
            return False
        
        c_atom_index = f_group_matches[0][1]
        
        amount_of_hydrogen = amount_of_hydrogen = Molecule3DFeaturesService._amount_of_hydrogen_in_neighbors(mol, c_atom_index)

        if amount_of_hydrogen != 2:
            return False
        
        return True
    
    def calculate_fluoric_group(self):
        """
        Identify the fluorine-containing functional group in the molecule.

        Returns:
            str: Identified functional group or 'non-F' if none found.
        """
        for f_group, f_group_SMILES in FUNCTIONAL_GROUP_TO_SMILES.items():
            f_group_submol = Chem.MolFromSmiles(f_group_SMILES)
            f_group_matches = self.mol.GetSubstructMatches(f_group_submol)

            if len(f_group_matches) > 0:
                if "gem-CF2" == f_group:
                    is_gem_CF2 = Molecule3DFeaturesService._check_gem_CF2(self.mol)
                    if not is_gem_CF2:
                        continue

                if "CH2F" == f_group:
                    is_CH2F = Molecule3DFeaturesService._check_CH2F(self.mol)
                    if not is_CH2F:
                        continue
                return f_group

        return "non-F"

    def calculate_f_group_freedom(self):
        """
        Calculate the freedom of the fluoric functional group.

        Args:
            functional_group (str): The functional group.

        Returns:
            int: The freedom of the functional group.
        """
        functional_group_to_freedom = {
                "CF3": 0, 
                "CH2F": 1, 
                "gem-CF2": 0, 
                "CHF2": 1,
                "CHF": 1,
                "non-F": 1,
            }
        
        return functional_group_to_freedom[self.f_group]

    def calculate_identificator(self) -> Identificator:
        """
        Identify the type of molecule.

        Returns:
            Identificator: Type of functional group.

        Raises:
            TypeError: If the molecule doesn't match any expected functional group.
        """
        if self.target_value == Target.pKa:
            return utils_pKa.calculate_identificator(self.mol)
        elif self.target_value == Target.logP:
            return utils_logP.calculate_identificator(self.mol)

    @staticmethod
    def set_average_atoms_position(mol,
                                   atoms_idx: list,
                                   conf_id: int = -1):
        """
        Create virtual atom in the molecule in the middle of atoms with atoms_idx.
        
        Args:
            mol: Rdkit optimized molecule.
            atoms_idx ([int]): Molecules indexes to find their average position.
            conf_id: Conformer id of the molecule. Defaults to -1.
            
        Returns:
            mol: Rdkit molecule with new virtual atom.
            idx (int): index of new virtual atom.
        """
        x, y, z = 0, 0, 0
        for atom_idx in atoms_idx:
            x += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[0]
            y += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[1]
            z += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[2]
        
        x /= len(atoms_idx)
        y /= len(atoms_idx)
        z /= len(atoms_idx)

        new_atom = rdchem.Atom(0)
        editable_molecule = RWMol(mol)
        idx = editable_molecule.AddAtom(new_atom)
        mol = editable_molecule.GetMol()

        conf = mol.GetConformer(conf_id)
        conf.SetAtomPosition(idx, Point3D(x,y,z))

        return mol, idx

    @staticmethod
    def change_vector_direction(mol, 
                                X1:int, R_1:int, 
                                conf_id:int):
        """
        Changes the direction of the vector X1R-1. Only executed when f_group is "secondary amine":
            1. We have to change direction of X1R-1 to (-1) * X1R1
            2. X1 is in the middle between R(-1) and R1, so R1 = 2*X1 - (R-1)
        
        Args:
            mol: Rdkit optimized molecule.
            X1 (int): Atom index that connects functional group(NH2 or COOH) to the cycle of the molecule.
            R_1 (int): Atom index of the center of the functional group(NH2 or COOH).
            conf_id (int): Conformer id of the molecule. Defaults to -1.
            
        Returns:
            mol: Rdkit molecule with new virtual atom.
            idx (int): index of new virtual atom.
        """ 
        conf = mol.GetConformer(conf_id)

        Rx_1 = conf.GetAtomPosition(R_1)[0]
        Ry_1 = conf.GetAtomPosition(R_1)[1]
        Rz_1 = conf.GetAtomPosition(R_1)[2]

        Xx1 = conf.GetAtomPosition(X1)[0]
        Xy1 = conf.GetAtomPosition(X1)[1]
        Xz1 = conf.GetAtomPosition(X1)[2]
        
        Rx1 = 2 * Xx1 - Rx_1
        Ry1 = 2 * Xy1 - Ry_1
        Rz1 = 2 * Xz1 - Rz_1

        conf.SetAtomPosition(R_1, Point3D(Rx1,Ry1,Rz1))
        return mol, R_1

    @staticmethod
    def dihedral_angle(mol, 
                       iAtomId:int, jAtomId:int, kAtomId:int, lAtomId:int, 
                       conf_id:int):
        """
        Calculate dihedral angle in molecule between IJK and JKL flat areas with specified conformer id.
        
        Args:
            mol: Rdkit optimized molecule.
            iAtomId (int): AtomI index in molecule.
            jAtomId (int): AtomJ index in molecule.
            kAtomId (int): AtomK index in molecule.
            lAtomId (int): AtomL index in molecule.
            conf_id: Conformer id of the molecule.
            
        Returns:
            (float): dihedral angle between areas in degrees.
        """
        conf = mol.GetConformer(conf_id)

        return abs(rdMolTransforms.GetDihedralDeg(conf, iAtomId, jAtomId, kAtomId, lAtomId))

    @staticmethod
    def is_atom_in_cycle(mol, atom_id):
        """
        Checks if atom is in cycle.
        
        Args:
            mol: Rdkit optimized molecule.
            atom_id (int): Atom id to check if its in cycle.
            
        Returns:
            is_atom_in_ring (bool): True if atom is in ring and False if not.
        """
        atom = mol.GetAtomWithIdx(atom_id)

        return atom.IsInRing()

    def calculate_volume(self):
        """
        Calculate the molecule Volume.

        Returns:
            mol_volume (float): Float value with calculated molecule volume
        """
        mol_volume = AllChem.ComputeMolVolume(mol=self.mol,
                                              confId=self.min_energy_conf_index)
        
        return mol_volume

    def calculate_sasa(self):
        """
        Calculate the molecule solvent accessible surface area.

        Returns:
            sasa (float): Float value with calculated molecule solvent accessible surface area.
        """
        mol_classify = rdFreeSASA.classifyAtoms(self.mol)
        sasa = rdFreeSASA.CalcSASA(mol=self.mol, 
                                   radii=mol_classify, 
                                   confIdx=self.min_energy_conf_index)
        
        return sasa
    
    def find_X1X2R1R2(self):
        """
        Determines which atoms correspond to X1, X2, R1 and R2
            
        Returns:
            X1 (int): Atom id in cycle, that connects NH2 or COOH to the molecule.
            X2 (int): Atom id in cycle, that connects fluorine functional group to the molecule.
            R1 (int): NH2 or COOH functional group center atom id.
            R2 (int): fluorine functional group center atom id.
        """
        f_group_smiles = FUNCTIONAL_GROUP_TO_SMILES[self.f_group]

        carboxile_submol = Chem.MolFromSmiles('CC=O')
        nitro_amine_submol = Chem.MolFromSmiles('CN')

        carboxile_matches = self.mol.GetSubstructMatches(carboxile_submol)
        nitro_amine_matches = self.mol.GetSubstructMatches(nitro_amine_submol)

        X1, R1 = None, None
        if self.identificator == Identificator.carboxilic_acid:
            if len(carboxile_matches) == 0:
                raise "Problem with carboxile acid"
            
            X1 = carboxile_matches[0][0]
            R1 = carboxile_matches[0][1]

        if "amine" in self.identificator.name.lower():
            if len(nitro_amine_matches) == 0:
                raise "Problem with amine"
            
            if Identificator.primary_amine == self.identificator:
                X1 = nitro_amine_matches[0][0]
                R1 = nitro_amine_matches[0][1]
            elif Identificator.secondary_amine == self.identificator:
                X1 = nitro_amine_matches[0][1]
                self.mol, R_1 = Molecule3DFeaturesService.set_average_atoms_position(self.mol, [nitro_amine_matches[0][0], nitro_amine_matches[1][0]], self.min_energy_conf_index)
                self.mol, R1 = Molecule3DFeaturesService.change_vector_direction(self.mol, X1, R_1=R_1, conf_id=self.min_energy_conf_index)

        X2, R2 = None, None
        f_group_submol = Chem.MolFromSmiles(f_group_smiles)
        f_group_matches = self.mol.GetSubstructMatches(f_group_submol)
        if self.f_group.upper() in ['CF3', 'CHF2', 'CH2F']:
            X2 = f_group_matches[0][0]
            R2 = f_group_matches[0][1]

        elif self.f_group == 'gem-CF2':
            X2 = f_group_matches[0][0]
            self.mol, R2 = Molecule3DFeaturesService.set_average_atoms_position(self.mol, [f_group_matches[0][1], f_group_matches[0][2]], self.min_energy_conf_index)

        elif self.f_group.upper() == 'CHF':
            if len(f_group_matches) == 1:
                X2 = f_group_matches[0][0]
                R2 = f_group_matches[0][1]
            elif len(f_group_matches) == 2:
                self.mol, X2 = Molecule3DFeaturesService.set_average_atoms_position(self.mol, [f_group_matches[0][0], f_group_matches[1][0]], self.min_energy_conf_index)
                self.mol, R2 = Molecule3DFeaturesService.set_average_atoms_position(self.mol, [f_group_matches[0][1], f_group_matches[1][1]], self.min_energy_conf_index)
        else:
            return X1, X2, R1, R2
        
        if len(set([X1, X2, R1, R2])) != 4:
            X1, X2, R1, R2 = None, None, None, None
        
        if not Molecule3DFeaturesService.is_atom_in_cycle(mol=self.mol, atom_id=f_group_matches[0][0]):
            X1, X2, R1, R2 = None, None, None, None

        return X1, X2, R1, R2

    def calculate_dihedral_angle(self):
        """
        Calculate dihedral angle in molecule between X1X2R1 and X2R1R2 flat areas with lowest energy conformer.
        
        Args:
            mol: Rdkit optimized molecule.
            iAtomId (int): AtomI index in molecule.
            jAtomId (int): AtomJ index in molecule.
            kAtomId (int): AtomK index in molecule.
            lAtomId (int): AtomL index in molecule.
            conf_id: Conformer id of the molecule.
            
        Returns:
            dihedral_angle_value (float): dihedral angle between areas.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            return None

        dihedral_angle_value = Molecule3DFeaturesService.dihedral_angle(self.mol, 
                                                                        self.R2, self.X2, self.X1, self.R1, 
                                                                        self.min_energy_conf_index) 
        return dihedral_angle_value
    
    @staticmethod
    def distance_between_atoms(iAtom_pos, jAtom_pos):
        """
        Calculate distance between two atoms.
        
        Args:
            iAtom_pos (tuple(int, int, int)): AtomI coordinates.
            jAtom_pos (tuple(int, int, int)): AtomJ coordinates.
            
        Returns:
            distance (float): distance between two points.
        """
        vector = (jAtom_pos[0] - iAtom_pos[0], jAtom_pos[1] - iAtom_pos[1], jAtom_pos[2] - iAtom_pos[2])

        distance = math.sqrt(pow(vector[0], 2) + pow(vector[1], 2) + pow(vector[2], 2))

        return distance

    def calculate_distance_between_atoms_in_cycle(self):
        """
        Calculate distance in molecule between X1 and X2.

        Returns:
            r_distance (float): dihedral between X1 and X2 atoms.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            return None

        X1_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.X1)
        X2_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.X2)

        r_distance = Molecule3DFeaturesService.distance_between_atoms(X1_pos, X2_pos)

        return r_distance
    
    def calculate_distance_between_atoms_in_f_group_centers(self):
        """
        Calculate distance in molecule between R1 and R2.

        Returns:
            R_distance (float): distance between R1 and R2 atoms.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            return None

        R1_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.R1)
        R2_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.R2)

        R_distance = Molecule3DFeaturesService.distance_between_atoms(R1_pos, R2_pos)

        return R_distance

    @staticmethod
    def flat_angle(mol, 
                   iAtomId:int, jAtomId:int, kAtomId:int, 
                   conf_id:int):
        """
        Calculate flat angle between I, J and K atoms in molecule.
        
        Args:
            iAtomId (int): AtomI coordinates.
            jAtomId (int): AtomJ coordinates.
            kAtomId (int): AtomK coordinates.
            
        Returns:
            (float): IJK Angle.
        """
        conf = mol.GetConformer(conf_id)

        return rdMolTransforms.GetAngleDeg(conf, iAtomId, jAtomId, kAtomId)

    def calculate_flat_angle_between_atoms_in_cycle(self):
        """
        Calculate 2 flat angles(in Deg) between 2 atoms in cycle and 1 atom in the center of functional group.

        If angle is not possible to calculate angles are None.

        Returns:
            flat_angle_between_atoms_in_cycle_1 (float): angle between X1, X2 and R2 atoms.
            flat_angle_between_atoms_in_cycle_2 (float): angle between X2, X1 and R1 atoms.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            return None, None

        flat_angle_between_atoms_in_cycle_1 = Molecule3DFeaturesService.flat_angle(self.mol, self.X1, self.X2, self.R2, self.min_energy_conf_index)
        flat_angle_between_atoms_in_cycle_2 = Molecule3DFeaturesService.flat_angle(self.mol, self.X2, self.X1, self.R1, self.min_energy_conf_index)

        return flat_angle_between_atoms_in_cycle_1, flat_angle_between_atoms_in_cycle_2

    def calculate_flat_angle_between_atoms_in_f_group_center(self):
        """
        Calculate 2 flat angles(in Deg) between 2 atoms in the center of functional groups and 1 atom in cycle.

        If angle is not possible to calculate angles are None.

        Returns:
            flat_angle_between_atoms_in_f_group_center_1 (float): angle between R2, X2 and R1 atoms.
            flat_angle_between_atoms_in_f_group_center_2 (float): angle between R1, X1 and R2 atoms.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            return None, None

        flat_angle_between_atoms_in_f_group_center_1 = Molecule3DFeaturesService.flat_angle(self.mol, self.R2, self.X2, self.R1, self.min_energy_conf_index)
        flat_angle_between_atoms_in_f_group_center_2 = Molecule3DFeaturesService.flat_angle(self.mol, self.R1, self.X1, self.R2, self.min_energy_conf_index)

        return flat_angle_between_atoms_in_f_group_center_1, flat_angle_between_atoms_in_f_group_center_2
    
    @staticmethod
    def _is_on_the_same_side(R1X1X2_angle: float,
                             R1X1R2_angle: float,
                             threshold: int = 5):
        """
        Determine if the functional group and fluorine substituents are on the same side of the molecule.
        For this R1X1X2 and R1X1R2 angles are used.

        Args:
            R1X1X2_angle (float): Angle R1X1X2.
            R1X1R2_angle (float): Angle R1X1R2.
            threshold (int, optional): Threshold value. Defaults to 5.

        Returns:
            bool: True if the functional group and fluorine substituents are on the same side, False otherwise.
        """
        if R1X1X2_angle - threshold > R1X1R2_angle:
            return True
        
        return False
    
    @staticmethod
    def _find_the_furthest_atom_id(mol: rdchem.Mol,
                                   atom_id: int):
        """
        Find the furthest atom ID from a given atom ID in a molecule using BFS.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            atom_id (int): Atom ID.

        Returns:
            tuple: Tuple containing the furthest atom ID and its distance.
        """
        queue = deque([(atom_id, 0)])
    
        visited = set()
        
        while queue:
            current_atom, distance = queue.popleft()
            
            visited.add(current_atom)
            
            neighbors = []
            for atom in mol.GetAtomWithIdx(current_atom).GetNeighbors():
                if atom.GetSymbol().lower() == 'h':
                    continue
                neighbors.append(atom.GetIdx())
            
            for neighbor in neighbors:
                if neighbor not in visited:
                    queue.append((neighbor, distance + 1))
        
        return current_atom, distance

    @staticmethod
    def _is_on_the_same_side_(mol: rdchem.Mol,
                              identificator: Identificator,
                              conf_id: int):
        """
        Determine if the functional group and fluorine substituents are on the same side of the molecule.
        For this R1X1X2 and R1X1R2 angles are used.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            identificator (Identificator): Identifier for specific atoms (acid, primary amine, secondary amine).
            conf_id (int): Conformer with the lowest energy.

        Returns:
            bool: True if the functional group and fluorine substituents are on the same side, False otherwise.
        """
        
        ring_submol = Chem.MolFromSmiles("C1=CC=CC=C1")
        ring_matches = mol.GetSubstructMatches(ring_submol)
        atoms_to_skip = ring_matches[0] if len(ring_matches) > 0 else []

        if identificator == Identificator.carboxilic_acid:
            # submol = Chem.MolFromSmiles('CC=O')
            submol = Chem.MolFromSmiles('C=O')
        else:
            submol = Chem.MolFromSmiles('CN')

        matches = mol.GetSubstructMatches(submol)
        atom_oxygen_idx = matches[0][1]
        # atomR1Idx = matches[0][1]
        # atomX1Idx = matches[0][0]
        atomX1Idx, atomR1Idx = utils.find_the_closest_atom_in_ring(mol=mol,
                                                                   atom_id=atom_oxygen_idx,
                                                                   atoms_not_to_visit=atoms_to_skip)

        atomR2Idx, _ = utils.find_the_furthest_atom(mol=mol, 
                                                    atom_id=atomR1Idx,
                                                    atoms_not_to_visit=atoms_to_skip)
        atomX2Neighbors = mol.GetAtomWithIdx(atomR2Idx).GetNeighbors()
        for atomX2Neighbor in atomX2Neighbors:
            if atomX2Neighbor.GetSymbol().lower() != 'h':
                atomX2Idx = atomX2Neighbor.GetIdx()
                break

        angle_R1X1R2 = Molecule3DFeaturesService.flat_angle(mol=mol,
                                                            iAtomId=atomR1Idx,
                                                            jAtomId=atomX1Idx,
                                                            kAtomId=atomR2Idx,
                                                            conf_id=conf_id)
        
        angle_R1X1X2 = Molecule3DFeaturesService.flat_angle(mol=mol,
                                                            iAtomId=atomR1Idx,
                                                            jAtomId=atomX1Idx,
                                                            kAtomId=atomX2Idx,
                                                            conf_id=conf_id)
        
        return Molecule3DFeaturesService._is_on_the_same_side(R1X1X2_angle=angle_R1X1X2,
                                                              R1X1R2_angle=angle_R1X1R2)

    @staticmethod
    def _amount_of_specific_atoms_in_molecule(mol: rdchem.Mol,
                                              atom_symbol: str = 'F'):
        """
        Calculate the number of specific atoms in a molecule.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            atom_symbol (str, optional): Atom symbol to count. Defaults to 'F'.

        Returns:
            int: Number of specific atoms in the molecule.
        """
        amount_of_atoms = 0
        
        for atom in mol.GetAtoms():
            if atom.GetSymbol().lower() == atom_symbol.lower():
                amount_of_atoms += 1

        return amount_of_atoms

    @staticmethod
    def _first_atoms_in_cycle(mol: rdchem.Mol,
                              atoms_id: list,
                              amount_of_atoms: int = 3):
        """
        Find the first atoms in a cycle starting from given atom IDs using BFS.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            atoms_id (list): List of starting atom IDs for BFS.
            amount_of_atoms (int, optional): Number of atoms in the cycle. Defaults to 3.

        Returns:
            list: List of atom IDs representing the first atoms in the cycle from given atom id.
        """
        queue = deque(atoms_id)
    
        visited = set()

        atoms_in_ring = []
        
        while queue:
            current_atom = queue.popleft()

            if mol.GetAtomWithIdx(current_atom).IsInRing() and current_atom not in atoms_in_ring:
                atoms_in_ring.append(current_atom)
                if len(atoms_in_ring) == amount_of_atoms:
                    return atoms_in_ring
            
            visited.add(current_atom)
            
            neighbors = []
            for atom in mol.GetAtomWithIdx(current_atom).GetNeighbors():
                if atom.GetSymbol().lower() == 'h':
                    continue
                neighbors.append(atom.GetIdx())
            
            for neighbor in neighbors:
                if neighbor not in visited:
                    queue.append(neighbor)
        
        return atoms_in_ring
    
    @staticmethod
    def _is_on_the_same_side_of_plane(plane_atoms_pos: list,
                                      atom1_pos: tuple,
                                      atom2_pos: tuple):
        """
        Determine if two atoms are on the same side of a plane.

        Args:
            plane_atoms_pos (list): Positions of atoms defining the plane.
            atom1_pos (tuple): Position of the first atom.
            atom2_pos (tuple): Position of the second atom.

        Returns:
            bool: True if atoms are on the same side of the plane, False otherwise.
        """
        atom_a = plane_atoms_pos[0]
        atom_b = plane_atoms_pos[1]
        atom_c = plane_atoms_pos[2]
        
        vector_ab = atom_b - atom_a
        vector_ac = atom_c - atom_a

        normal_vector = np.cross(vector_ab, vector_ac)

        d = -np.dot(normal_vector, atom_a)

        plane_eq_value_f_1 = np.dot(atom1_pos, normal_vector) + d
        plane_eq_value_f_2 = np.dot(atom2_pos, normal_vector) + d

        if np.sign(plane_eq_value_f_1) == np.sign(plane_eq_value_f_2):
            return True
        else:
            return False

    @staticmethod
    def _is_R1_R2_on_the_same_side(mol: rdchem.Mol,
                                   R1: int,
                                   X1: int,
                                   conf_id: int = -1):
        """
        Determine if two atoms are on the same side of a plane defined by a ring.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            R1 (int): Atom R1 ID.
            X1 (int): Atom X1 ID.
            conf_id (int, optional): Conformer id with the lowest energy. Defaults to -1.

        Returns:
            bool: True if atoms R1 and R2 are on the same side of the plane, False otherwise.
        """
        for atom in mol.GetAtoms():
            if atom.GetSymbol().lower() == 'f':
                fluorine_atom = atom.GetIdx()
                break

        first_atom_from_fluorine_in_ring = Molecule3DFeaturesService._first_atoms_in_cycle(mol=mol,
                                                                                           atoms_id=[fluorine_atom],
                                                                                           amount_of_atoms=1)[0]
        
        X2_atom = mol.GetAtomWithIdx(first_atom_from_fluorine_in_ring)
        for neighbor in X2_atom.GetNeighbors():
            if not neighbor.IsInRing() and neighbor.GetSymbol().lower() != 'h':
                R2_atom = neighbor
                break

        X2 = X2_atom.GetIdx()
        R2 = R2_atom.GetIdx()

        three_atoms_in_ring = Molecule3DFeaturesService._first_atoms_in_cycle(mol=mol,
                                                                              atoms_id=[X1],
                                                                              amount_of_atoms=3)

        atoms_from__ring_pos = []
        for atom_in_ring in three_atoms_in_ring:
            atoms_from__ring_pos.append(np.array(mol.GetConformer(conf_id).GetAtomPosition(atom_in_ring)))

        atom_1 = np.array(mol.GetConformer(conf_id).GetAtomPosition(R1))
        atom_2 = np.array(mol.GetConformer(conf_id).GetAtomPosition(R2))

        return Molecule3DFeaturesService._is_on_the_same_side_of_plane(plane_atoms_pos=atoms_from__ring_pos, 
                                                                       atom1_pos=atom_1,
                                                                       atom2_pos=atom_2)

    @staticmethod
    def _is_fluorines_on_the_same_side(mol: rdchem.Mol,
                                       conf_id: int = -1):
        """
        Determine if two fluorine atoms are on the same side of a plane.
        Used for molecules where we have 2 CHF groups.

        Args:
            mol (rdchem.Mol): RDKit molecule.
            conf_id (int, optional): Conformer id with the lowest energy. Defaults to -1.

        Returns:
            bool: True if fluorine atoms are on the same side of the plane, False otherwise.
        """
        fluorines = []

        for atom in mol.GetAtoms():
            if atom.GetSymbol().lower() == 'f':
                fluorines.append(atom.GetIdx())

        fluorine_atom_1 = fluorines[0]
        fluorine_atom_2 = fluorines[1]

        atoms_from_fluorine_in_ring = Molecule3DFeaturesService._first_atoms_in_cycle(mol=mol,
                                                                                      atoms_id=[fluorine_atom_1, fluorine_atom_2],
                                                                                      amount_of_atoms=3)

        atoms_from_fluorine_in_ring_pos = []
        for atom_from_fluorine_in_ring in atoms_from_fluorine_in_ring:
            atoms_from_fluorine_in_ring_pos.append(np.array(mol.GetConformer(conf_id).GetAtomPosition(atom_from_fluorine_in_ring)))

        atom_f_1 = np.array(mol.GetConformer(conf_id).GetAtomPosition(fluorine_atom_1))
        atom_f_2 = np.array(mol.GetConformer(conf_id).GetAtomPosition(fluorine_atom_2))

        return Molecule3DFeaturesService._is_on_the_same_side_of_plane(plane_atoms_pos=atoms_from_fluorine_in_ring_pos, 
                                                                       atom1_pos=atom_f_1,
                                                                       atom2_pos=atom_f_2)
        
    def calculate_cis_trans(self):
        """
        Calculate cis-trans configuration based on specific conditions for rdkit molecule.

        - if there is no chiral centers in the molecule or no "@" in the molecule cis/trans is nan.
        - if We haven't any fluorine groups in our molecule we are using COOH or NH2 and the furthest atom from it.
            For defining cis/trans we are checking two angles from functional group to the furthest atom.
        - If we have 2 CHF groups we are checking if fluorine atoms are the from the same part of molecule or they are
            from the different part of the molecule.
        - In all others cases we are looking to X1R1X2, X1R1R2 angles and their values, 
            if X1R1X2 > X1R1R2 then molecules are on the same part.
        
        Returns:
            'cis' if atoms are on the same side, 'trans' otherwise, nan if there is no chiral centers in the molecule.
        """
        if "@" not in self.smiles:
            return np.nan
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            if self.f_group == "non-F":
                is_on_the_same_side = Molecule3DFeaturesService._is_on_the_same_side_(mol=self.mol,
                                                                                      identificator=self.identificator,
                                                                                      conf_id=self.min_energy_conf_index) 
        else:
            if self.f_group == "CHF" and Molecule3DFeaturesService._amount_of_specific_atoms_in_molecule(mol=self.mol,
                                                                                                         atom_symbol='F') == 2:
                is_on_the_same_side = Molecule3DFeaturesService._is_fluorines_on_the_same_side(mol=self.mol,
                                                                                               conf_id=self.min_energy_conf_index)
            else:
                is_on_the_same_side = Molecule3DFeaturesService._is_on_the_same_side(self.flat_angle_between_atoms_in_cycle_2,
                                                                                     self.flat_angle_between_atoms_in_f_group_center_2)
        if is_on_the_same_side:
            return "cis"
        else:
            return "trans"
        
    def calculate_linear_path_f_to_fg(self):
        """
        Calculate amount of linear paths from fluorine (F) to the functional group (FG) based on the target value.

        If the target value is pKa, calculate the linear path from fluorine to the functional group using the provided SMILES.
        If the target value is logP, calculate the linear path from fluorine to the functional group using the provided SMILES and identificator.

        Returns:
            int: Amount of the linear paths from fluorine to the functional group.

        Raises:
            ValueError: If the molecule identificator is inappropriate.
        """
        if self.target_value == Target.pKa:
            return utils_pKa.calculate_linear_path_f_to_fg(smiles=self.smiles)
        elif self.target_value == Target.logP:
            return utils_logP.calculate_linear_path_f_to_fg(smiles=self.smiles,
                                                            identificator=self.identificator)
        
        raise ValueError("inappropriate molecule identificator")
                       
    def calculate_dipole_moment(self):
        """
        Calculate the dipole moment based on atoms positions and charges using optimized molecule with lowest energy conformer.

        Returns:
            dipole_moment (float): Float value with calculated dipole_moment
        """
        charges = []
        coordinates = []
        x_centroid, y_centroid, z_centroid = 0, 0, 0
        for atom in self.mol.GetAtoms():
            pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(atom.GetIdx())
            charge = atom.GetDoubleProp("_GasteigerCharge")

            charges.append(charge)
            coordinates.append([pos[0], pos[1], pos[2]])

            x_centroid += pos[0]
            y_centroid += pos[1]
            z_centroid += pos[2]

        x_centroid /= len(self.mol.GetAtoms())
        y_centroid /= len(self.mol.GetAtoms())
        z_centroid /= len(self.mol.GetAtoms())

        charges_multiply_coordinates = coordinates.copy()
        for charges_multiply_coordinate_index in range(len(charges_multiply_coordinates)):
            charges_multiply_coordinates[charges_multiply_coordinate_index][0] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][0] - x_centroid) * charges[charges_multiply_coordinate_index]
            charges_multiply_coordinates[charges_multiply_coordinate_index][1] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][1] - y_centroid) * charges[charges_multiply_coordinate_index]
            charges_multiply_coordinates[charges_multiply_coordinate_index][2] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][2] - z_centroid) * charges[charges_multiply_coordinate_index]

        dipole_moment_vector = [0, 0, 0]
        for charges_multiply_coordinate_index in range(len(charges_multiply_coordinates)):
            dipole_moment_vector[0] += charges_multiply_coordinates[charges_multiply_coordinate_index][0]
            dipole_moment_vector[1] += charges_multiply_coordinates[charges_multiply_coordinate_index][1]
            dipole_moment_vector[2] += charges_multiply_coordinates[charges_multiply_coordinate_index][2]

        dipole_moment = math.sqrt(pow(dipole_moment_vector[0], 2) + pow(dipole_moment_vector[1], 2) + pow(dipole_moment_vector[2], 2))

        return dipole_moment
    
    def calculate_molecular_weight(self):
        """
        Calculate the molecular weight based on the target value.

        If the target value is pKa, the molecular weight is calculated using only the SMILES provided.
        If the target value is logP, the molecular weight is calculated using the SMILES and identificator.

        Returns:
            float: The calculated molecular weight, rounded to three decimal places.
        """
        if self.target_value == Target.pKa:
            molecular_weight =  utils_pKa.calculate_molecular_weight(SMILES=self.smiles)
        elif self.target_value == Target.logP:
            molecular_weight = utils_logP.calculate_molecular_weight(SMILES=self.smiles,
                                                         identificator=self.identificator)
        return round(molecular_weight, 3)

    def calculate_TPSA_with_fluor(self):
        """
        Calculate the molecule topological polar surface area with additional fluorine area.

        Returns:
            tpsa_f (float): Float value with calculated molecule topological polar surface area 
            with additional fluorine area.
        """
        tpsa = Descriptors.TPSA(self.mol)
        fluor_idxs = [atom.GetIdx() for atom in self.mol.GetAtoms() if atom.GetSymbol().lower() == 'f']
        tpsa_f = tpsa

        radii = rdFreeSASA.classifyAtoms(self.mol)
        rdFreeSASA.CalcSASA(self.mol, radii)
        
        for fluor_idx in fluor_idxs:
            atom_sasa = self.mol.GetAtoms()[fluor_idx].GetProp('SASA')
            tpsa_f += float(atom_sasa)

        return tpsa_f



from enum import Enum
import numpy as np

class Target(Enum):
    pKa = 'pKa'
    logP = 'logP'

class Identificator(Enum):
    carboxilic_acid = 'carboxilic_acid'
    primary_amine = 'primary_amine'
    secondary_amine = 'secondary_amine'

LOGP_FEATURES = ['f_freedom', 'PPSA5', 'mol_num_cycles', 'nFRing', 'nF', 'identificator',
                 'mol_weight', 'dipole_moment', 'nHRing', 'nO', 'PBF', 'nC', 'nARing',
                 'cis/trans', 'PNSA5', 'FPSA3', 'mol_volume', 'RPCS', 'GeomShapeIndex',
                 'WPSA5', 'TASA', 'f_to_fg', 'avg_atoms_in_cycle', 'nFHRing',
                 'chirality']

LOGP_MODEL_PATH = r'fluoricLogPpKa\ml_part\models_weights\h2o_models\logP\all_molecules(without_angle_feature)_without_outliers\StackedEnsemble_BestOfFamily_3_AutoML_2_20240208_214951'

PKA_FEATURES = ['RPCS', 'PBF', 'mol_weight', 'dipole_moment', 'PPSA5',
                'avg_atoms_in_cycle', 'nHRing', 'cis/trans', 'FPSA3', 'nF', 'chirality',
                'sasa', 'PNSA5', 'GeomShapeIndex', 'TASA', 'mol_num_cycles',
                'f_freedom', 'nFRing', 'identificator', 'nO', 'nARing', 'nC', 'nFHRing',
                'f_to_fg']

PKA_AMINE_MODEL_PATH = r'fluoricLogPpKa\ml_part\models_weights\h2o_models\pKa\amine_molecules(without_angle_feature)_without_outliers\StackedEnsemble_BestOfFamily_5_AutoML_3_20240213_92029'
PKA_ACID_MODEL_PATH = r'fluoricLogPpKa\ml_part\models_weights\h2o_models\pKa\acid_molecules(without_angle_feature)_without_outliers\DeepLearning_grid_2_AutoML_4_20240213_102321_model_43'

FUNCTIONAL_GROUP_TO_SMILES = {
            "CF3": "CC(F)(F)F", 
            "CH2F": "CCF", 
            "gem-CF2": "C(F)(F)", 
            "CHF2": "CC(F)(F)",
            "CHF": "CF",
            "non-F": ""
        }

CONVERT_FEATURE_TO = {
    "identificator": {
        Identificator.carboxilic_acid: 0,
        Identificator.primary_amine: 1,
        Identificator.secondary_amine: 2,
    },
    "cis/trans": {
        np.nan: 0,
        "": 0,
        "cis": 1,
        "trans": 2
    }
}

ALL_SUBMOLS = {
    "CF3CH2N": ["FC(CN)(F)F"],
    "CF3(CH2)2N": ["FC(CCN)(F)F"],
    "CF3(CH2)3N": ["FC(CCCN)(F)F"],
    "CF3(CH2)4N": ["FC(CCCCN)(F)F"],
    "CF3(CH2)5N": ["FC(CCCCCN)(F)F"],
    "CHF2CH2N": ["FC(CN)([H])F"],
    "CHF2(CH2)2N": ["FC(CCN)([H])F"],
    "CHF2(CH2)3N": ["FC(CCCN)([H])F"],
    "CHF2(CH2)4N": ["FC(CCCCN)([H])F"],
    "CHF2(CH2)5N": ["FC(CCCCCN)([H])F"],
    "CH2FCH2N": ["FC(CN)([H])[H]"],
    "CH2F(CH2)2N": ["FC(CCN)([H])[H]"],
    "CH2F(CH2)3N": ["FC(CCCN)([H])[H]"],
    "CH2F(CH2)4N": ["FC(CCCCN)([H])[H]"],
    "CH2F(CH2)5N": ["FC(CCCCCN)([H])[H]"],
    "CCF2CN": ["FC(CN)(F)C"],
    "CCF2CCN": ["FC(CCN)(F)C", "FC(F)(C1)C1N"],
    "CCF2CCCN": ["FC(CCCN)(F)C", "FC(F)(CC1)C1N"],
    "CCF2CCCCN": ["FC(CCCCN)(F)C", "FC(F)(CCC1)C1N"],
    "CCF2CCCCCN": ["FC(CCCCCN)(F)C", "FC(F)(CCCC1)C1N"],
    "CCFHCN": ["FC(CN)([H])C"],
    "CCFHCCN": ["FC(CCN)([H])C"],
    "CCFHCCCN": ["FC(CCCN)([H])C"],
    "CCFHCCCCN": ["FC(CCCCN)([H])C"],
    "CCFHCCCCCN": ["FC(CCCCCN)([H])C"],
    "CF3CH2COOH": ["FC(CC(O)=O)(F)F"],
    "CF3(CH2)2COOH": ["FC(CCC(O)=O)(F)F"],
    "CF3(CH2)3COOH": ["FC(CCCC(O)=O)(F)F"],
    "CF3(CH2)4COOH": ["FC(CCCCC(O)=O)(F)F"],
    "CF3(CH2)5COOH": ["FC(CCCCCC(O)=O)(F)F"],
    "CHF2CH2COOH": ["FC(CC(O)=O)([H])F"],
    "CHF2(CH2)2COOH": ["FC(CCC(O)=O)([H])F"],
    "CHF2(CH2)3COOH": ["FC(CCCC(O)=O)([H])F"],
    "CHF2(CH2)4COOH": ["FC(CCCCC(O)=O)([H])F"],
    "CHF2(CH2)5COOH": ["FC(CCCCCC(O)=O)([H])F"],
    "CH2FCH2COOH": ["FC(CC(O)=O)([H])[H]"],
    "CH2F(CH2)2COOH": ["FC(CCC(O)=O)([H])[H]"],
    "CH2F(CH2)3COOH": ["FC(CCCC(O)=O)([H])[H]"],
    "CH2F(CH2)4COOH": ["FC(CCCCC(O)=O)([H])[H]"],
    "CH2F(CH2)5COOH": ["FC(CCCCCC(O)=O)([H])[H]"],
    "CCF2CCOOH": ["FC(CC(O)=O)(F)C"],
    "CCF2CCCOOH": ["FC(CCC(O)=O)(F)C", "FC1(F)CC1C(O)=O"],
    "CCF2CCCCOOH": ["FC(CCCC(O)=O)(F)C", "FC1(F)CCC1C(O)=O"],
    "CCF2CCCCCOOH": ["FC(CCCCC(O)=O)(F)C", "FC1(F)CCCC1C(O)=O"],
    "CCF2CCCCCCOOH": ["FC(CCCCCC(O)=O)(F)C", "FC1(F)CCCCC1C(O)=O"],
    "CCFHCCOOH": ["FC(CC(O)=O)([H])C"],
    "CCFHCCCOOH": ["FC(CCC(O)=O)([H])C"],
    "CCFHCCCCOOH": ["FC(CCCC(O)=O)([H])C"],
    "CCFHCCCCCOOH": ["FC(CCCCC(O)=O)([H])C"],
    "CCFHCCCCCCOOH": ["FC(CCCCCC(O)=O)([H])C"]
}


import pandas as pd

from fluoriclogppka.ml_part.constants import Target, Identificator
from fluoriclogppka.ml_part.constants import LOGP_MODEL_PATH, PKA_AMINE_MODEL_PATH, PKA_ACID_MODEL_PATH

from fluoriclogppka.ml_part.data_preparation.smiles_to_features import Featurizer
from fluoriclogppka.ml_part.services.h2o_service import H2OService

class Inference:
    """
    A class for making predictions using pre-trained models based on input molecule information.

    This class prepares data for prediction and utilizes an H2OService to make predictions
    using the appropriate pre-trained model.

    Attributes:
        SMILES (str): The SMILES string representing the molecule.
        target_value (Target): The target property to predict (pKa or logP).
        row_from_enamine_dataset: A row from the Enamine dataset containing molecule information.
        model_path (str): The path to the pre-trained model.

    Methods:
        best_model_path(): Determines the best model path based on the target value and molecule type.
        predict(): Makes predictions using the loaded model.
    """
    def __init__(self, 
                 SMILES: str,
                 target_value: Target = Target.pKa,
                 model_path: str = None
                 ) -> None:
        """
        Initialize the Inference object.

        Args:
            SMILES (str): The SMILES string representing the molecule.
            target_value (Target): The target property to predict (pKa or logP).
            row_from_enamine_dataset: A row from the Enamine dataset containing molecule information.
            model_path (str, optional): The path to the pre-trained model. Defaults to None.
        """
        dataPrep = Featurizer(SMILES=SMILES,
                              target_value=target_value,)
        self.features_for_predict = dataPrep.features_for_predict

        identificator = dataPrep.all_features_dict['identificator']

        self.model_path = model_path
        if model_path is None:
            self.model_path = Inference.best_model_path(target_value=target_value,
                                                        identificator=identificator)
        
        self._h2oService = H2OService(self.model_path)
        self.model = self._h2oService.model
    
    @staticmethod
    def best_model_path(target_value: Target, identificator: Identificator):
        """
        Determines the best model path based on the target value and molecule type.

        Args:
            target_value (Target): The target property to predict (pKa or logP).
            identificator (Identificator): The type of the molecule.

        Returns:
            str: The path to the pre-trained model.
        """
        if target_value == Target.logP:
            model_path = LOGP_MODEL_PATH
        elif target_value == Target.pKa and "amine" in identificator.name.lower():
            model_path = PKA_AMINE_MODEL_PATH
        elif target_value == Target.pKa and "acid" in identificator.name.lower():
            model_path = PKA_ACID_MODEL_PATH
        
        return model_path

    def predict(self):
        """
        Make predictions using the loaded model.

        Returns:
            predicted_value (float): Predicted pKa or logP target value.
        """
        predicted_value = self._h2oService.predict(self.features_for_predict)
        
        return predicted_value


if __name__ == "__main__":
    SMILES = "F[C@H]1C[C@H](F)CN(C1)C(=O)C1=CC=CC=C1"
    
    inference = Inference(SMILES=SMILES,
                          target_value=Target.logP)
    
    predicted_pKa = inference.predict()
    print(predicted_pKa)



from rdkit import Chem
from rdkit.Chem import Descriptors

from fluoriclogppka.ml_part.constants import Identificator
from fluoriclogppka.ml_part.constants import ALL_SUBMOLS
from fluoriclogppka.ml_part.services.utils import find_the_furthest_atom, find_all_atoms_from

def calculate_identificator(mol) -> Identificator:
    """
    Identify the type of functional group present in the molecule - COOH, NH or NH2.

    Returns:
        Identificator: Type of functional group.

    Raises:
        TypeError: If the molecule doesn't match any expected functional group.
    """
    ring_submol = Chem.MolFromSmiles("C1=CC=CC=C1")
    COOH_submol = Chem.MolFromSmiles("C(=O)")
    NH_submol = Chem.MolFromSmiles("N")
    secondary_amine_submol = Chem.MolFromSmiles("CN(C)C")

    ring_matches = mol.GetSubstructMatches(ring_submol)
    COOH_matches = mol.GetSubstructMatches(COOH_submol)
    NH_matches = mol.GetSubstructMatches(NH_submol)
    secondary_amine_matches = mol.GetSubstructMatches(secondary_amine_submol)

    furthest_atom_from_COOH, furthest_COOH_distance = find_the_furthest_atom(mol=mol,
                                                                             atom_id=COOH_matches[0][0],
                                                                             atoms_not_to_visit=ring_matches[0])
    
    furthest_atom_from_NH, furthest_NH_distance = find_the_furthest_atom(mol=mol,
                                                                         atom_id=NH_matches[0][0],
                                                                         atoms_not_to_visit=ring_matches[0])
    
    if furthest_NH_distance > furthest_COOH_distance:
        return Identificator.carboxilic_acid
    
    if len(secondary_amine_matches) > 0:
        return Identificator.secondary_amine
    else:
        return Identificator.primary_amine
    
def calculate_linear_path_f_to_fg(smiles: str,
                                  identificator: Identificator):
    """
    Calculate the amount of linear paths from fluorine (F) to the functional group (FG) based on the provided SMILES 
    and molecule identificator.

    This function constructs a molecule from the given SMILES representation and identifies the linear path 
    from fluorine to the specified functional group. It iterates over all substructures associated with the 
    functional group and counts the total number of matches found in the molecule, excluding certain substructures 
    based on the provided identificator.

    Args:
        smiles (str): SMILES representation of the molecule.
        identificator (Identificator): Identificator specifying the type of functional group.

    Returns:
        int: The total number of matches found between substructures and the molecule.
    """
    mol = Chem.MolFromSmiles(smiles, sanitize=True) 
    mol = Chem.AddHs(mol) 
    
    total_matches = 0
    for _, sub_SMILES_array in ALL_SUBMOLS.items():
        for sub_SMILES in sub_SMILES_array:
            
            if identificator == Identificator.carboxilic_acid and "n" in sub_SMILES.lower():
                continue
            elif "amine" in identificator.name.lower() and "(O)=O" in sub_SMILES:
                continue

            sub_SMILES = sub_SMILES.replace("(O)=", "(N)=")
            
            submol = Chem.MolFromSmiles(sub_SMILES, sanitize=False)
            matches = mol.GetSubstructMatches(submol)

            total_matches += len(matches)
    
    return total_matches

def calculate_molecular_weight(SMILES: str,
                               identificator: Identificator):
    """
    Calculate the molecular weight of the molecule based on the provided SMILES and molecule identificator.

    This function calculates the molecular weight of the molecule represented by the given SMILES. 
    It first constructs the molecule from the SMILES, then identifies specific substructures relevant to 
    the provided identificator (such as carbonyl groups for carboxylic acids or amine groups for amines). 
    After removing unnecessary atoms, it sanitizes the molecule, replaces certain substructures, adds hydrogens, 
    and finally calculates the molecular weight.

    Args:
        SMILES (str): SMILES representation of the molecule.
        identificator (Identificator): Identificator specifying the type of functional group.

    Returns:
        float: The calculated molecular weight.
    
    Raises:
        TypeError: If the type of molecule is inappropriate for the given identificator.
    """
    mol = Chem.MolFromSmiles(SMILES)

    CO_submol = Chem.MolFromSmiles("C=O")
    NH_submol = Chem.MolFromSmiles("N")

    CO_matches = mol.GetSubstructMatches(CO_submol)
    NH_matches = mol.GetSubstructMatches(NH_submol)

    if len(NH_matches) == 0 or len(CO_matches) == 0:
        raise TypeError("Inappropriate type of the molecule")

    if identificator == Identificator.carboxilic_acid:
        atoms_to_remove = find_all_atoms_from(mol=mol, 
                                            from_atom_id=NH_matches[0][0], 
                                            atoms_not_to_visit=CO_matches[0])
    else:
        atoms_to_remove = find_all_atoms_from(mol=mol, 
                                            from_atom_id=CO_matches[0][0], 
                                            atoms_not_to_visit=NH_matches[0])
        
    rwmol = Chem.RWMol(mol) 
    for atom_idx in reversed(list(atoms_to_remove)):
        rwmol.RemoveAtom(atom_idx)
    mol = rwmol.GetMol()

    Chem.SanitizeMol(mol)

    mol = Chem.ReplaceSubstructs(mol, 
                                Chem.MolFromSmiles('C=O'), 
                                Chem.MolFromSmiles('C(=O)O'),
                                replaceAll=True)[0]
    mol = Chem.AddHs(mol)

    return Descriptors.MolWt(mol)



from collections import deque

from rdkit import Chem
from rdkit.Chem import rdchem

def has_numbers(inputString: str):
    """
    Checks if there is a number in the string.

    Args:
        - inputString (str): Text, that needs to be checked. Example: "acb1"
    
    Returns:
        - (bool): True if there is a number in the function, False if not.
    """
    return any(char.isdigit() for char in inputString)

def cycles_amount(mol):
    """
        Calculate amount of cycles in the molecule.
        
        Args:
            - mol: rdkit molecule.
            
        Returns:
            - amount_of_cycles (int): amount of rings in the molecule.
        """
    sssr = Chem.GetSSSR(mol)
    num_rings = len(sssr)
    return num_rings

def find_the_furthest_atom(mol: rdchem.Mol, 
                           atom_id: int, 
                           atoms_not_to_visit: list = []):
    """
    Find the furthest atom from a given atom ID in a molecule, excluding specified atoms.

    This function performs a BFS starting from the provided atom ID to find the furthest 
    atom in the molecule. It iterates through the neighbors of each atom, excluding hydrogen atoms and 
    those specified in the 'atoms_not_to_visit' list, and tracks the distance from the starting atom. 
    The search continues until no more unvisited neighbors remain, and the furthest atom and its distance 
    from the starting atom are returned.

    Args:
        mol (rdchem.Mol): RDKit molecule.
        atom_id (int): Index of the starting atom.
        atoms_not_to_visit (list, optional): List of atom indices to exclude from the search. Defaults to [].

    Returns:
        tuple: Tuple containing the index of the furthest atom and its distance from the starting atom.
    """
    queue = deque([(atom_id, 0)])

    visited = set()
    
    while queue:
        current_atom, distance = queue.popleft()
        
        visited.add(current_atom)
        
        neighbors = []
        for atom in mol.GetAtomWithIdx(current_atom).GetNeighbors():
            if atom.GetSymbol().lower() == 'h':
                continue
            if atom.GetIdx() in atoms_not_to_visit:
                continue
            neighbors.append(atom.GetIdx())
        
        for neighbor in neighbors:
            if neighbor not in visited:
                queue.append((neighbor, distance + 1))
    
    return current_atom, distance

def find_the_closest_atom_in_ring(mol: rdchem.Mol, 
                                  atom_id: int, 
                                  atoms_not_to_visit: list = []):
    """
    Find the closest atom within a ring from a given atom ID in a molecule, excluding specified atoms.

    This function performs BFS starting from the provided atom ID to find the closest 
    atom within a ring in the molecule. It iterates through the neighbors of each atom, excluding hydrogen 
    atoms and those specified in the 'atoms_not_to_visit' list, until a ring atom is encountered. 
    The search stops at the first encountered ring atom, and its index along with the index of the previous 
    atom in the search path are returned.

    Args:
        mol (rdchem.Mol): RDKit molecule.
        atom_id (int): Index of the starting atom.
        atoms_not_to_visit (list, optional): List of atom indices to exclude from the search. Defaults to [].

    Returns:
        tuple: Tuple containing the index of the closest ring atom and the index of the previous atom in the search path.
    """
    queue = deque([(atom_id, -1)])

    visited = set()
    
    while queue:
        current_atom, previous_atom = queue.popleft()
        if mol.GetAtomWithIdx(current_atom).IsInRing():
            break
        visited.add(current_atom)
        
        neighbors = []
        for atom in mol.GetAtomWithIdx(current_atom).GetNeighbors():
            if atom.GetSymbol().lower() == 'h':
                continue
            if atom.GetIdx() in atoms_not_to_visit:
                continue
            neighbors.append(atom.GetIdx())
        
        for neighbor in neighbors:
            if neighbor not in visited:
                queue.append((neighbor, current_atom))
    
    return current_atom, previous_atom

def find_all_atoms_from(mol: rdchem.Mol, 
                        from_atom_id: int, 
                        atoms_not_to_visit: list = []):
    """
    Find all atoms reachable from a given starting atom ID in a molecule, excluding specified atoms.

    This function performs BFS starting from the provided starting atom ID to find 
    all reachable atoms in the molecule. It iterates through the neighbors of each atom, excluding hydrogen 
    atoms and those specified in the 'atoms_not_to_visit' list, and adds unvisited neighbors to the queue 
    for further exploration. The search continues until all reachable atoms are visited, and a set containing 
    the indices of all visited atoms is returned.

    Args:
        mol (rdchem.Mol): RDKit molecule.
        from_atom_id (int): Index of the starting atom.
        atoms_not_to_visit (list, optional): List of atom indices to exclude from the search. Defaults to [].

    Returns:
        set: Set containing the indices of all atoms reachable from the starting atom.
    """
    queue = deque([(from_atom_id, -1)])

    visited = set()
    
    while queue:
        current_atom, previous_atom = queue.popleft()

        visited.add(current_atom)
        
        neighbors = []
        for atom in mol.GetAtomWithIdx(current_atom).GetNeighbors():
            if atom.GetSymbol().lower() == 'h':
                continue
            if atom.GetIdx() in atoms_not_to_visit:
                continue
            neighbors.append(atom.GetIdx())
        
        for neighbor in neighbors:
            if neighbor not in visited:
                queue.append((neighbor, current_atom))
    
    return visited



__all__ = [
    "DipoleMoment",
    "MoleculeVolume",
    "MoleculeSASA",
    "MoleculeTPSAF",
    "MoleculeDihedralAngle",
    "DistanceBetweenX1X2",
    "DistanceBetweenR1R2",
    "AngleX1X2R2",
    "AngleX2X1R1",
    "AngleR2X2R1",
    "AngleR1X1R2",
    "MoleculeRingsAmount",
    "AtomsToRingRatio",
    "Chirality"
]

import math

from rdkit import Chem
from rdkit.Chem import AllChem, Descriptors
from rdkit.Chem import rdForceFieldHelpers, rdPartialCharges
from rdkit.Chem import rdFreeSASA, rdchem, rdMolTransforms
from rdkit.Chem.rdchem import RWMol
from rdkit.Geometry import Point3D

from fluoriclogppka.ml_part.constants import Identificator
from fluoriclogppka.ml_part.constants import FUNCTIONAL_GROUP_TO_SMILES
from fluoriclogppka.ml_part.exceptions import InvalidMoleculeTypeError
from fluoriclogppka.ml_part.services.utils import cycles_amount

class OptimizedMolecule:
    """
    Class that represents a geometrically optimized molecule.

    This class provides functionality to obtain optimized molecule, 
    its conformer with lowest energy and atoms indexes for calculating 3D features.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        functional_group_to_smiles (dict): Dict to convert string f_group to smiles. Defaults to None
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        find_conf_with_min_energy(): Get the conformer index with minimal energy.
        prepare_molecule(): Creates 3D sanitized molecule with multiple conformers from smiles.
        set_average_atoms_position(): Creates virtual atom in the middle of given atoms.
        change_vector_direction(): Changes the direction of the atoms vector.
        is_atom_in_cycle(): Checks if atom is in cycle.
        find_X1X2R1R2(): Determines atoms indexes to X1, X2, R1, R2.
        optimize_geometry(): Optimizes molecule's conformers and returns with min energy.
    """
    def __init__(self, 
                 smiles: str,
                 f_group: str = None,
                 identificator: Identificator = None):
        """
        Initialize the OptimizedMolecule instance and calculates atoms indexes for 3d features.
        If fluor functional groups or identificator is None X1, X2, R1, R2 cannot be calculated.

        Args:
            smiles (str): String representation of a molecule.
            f_group (str): Fluor functional group name. Defaults to None.
            identificator (Identificator): The molecule type. Defaults to None.
        """
        self.smiles = smiles
        self.f_group = f_group
        self.functional_group_to_smiles = FUNCTIONAL_GROUP_TO_SMILES
        self.identificator = identificator

        self.optimize_geometry()

        if f_group is not None and identificator is not None:
            self.X1, self.X2, self.R1, self.R2 = self.find_X1X2R1R2()

    @staticmethod
    def find_conf_with_min_energy(mol):
        """
        Optimizes all molecules conformers and finds the one with the lowest energy.
        
        Args:
            mol: 3D sanitized molecule with multiple conformers. 

        Returns:
            min_energy_conf_index (int): Conformer index with lowest energy.
            min_energy (float): The lowest conformer energy of the optimized molecule.
            mol: Rdkit optimized molecule.
        """
        optimization_result = rdForceFieldHelpers.MMFFOptimizeMoleculeConfs(mol)
        
        min_energy, min_energy_conf_index = pow(10,5), None
        for index, (status, energy) in enumerate(optimization_result):
            if energy < min_energy and status == 0:
                min_energy_conf_index = index
                min_energy = min(min_energy, energy)

        return min_energy_conf_index, min_energy, mol
    
    @staticmethod
    def prepare_molecule(smiles):
        """
        Create rdkit 3d molecule from SMILES.
        Generate charges and molecule's conformers.
        
        Args:
            smiles (str): String representation of a molecule. 
            
        Returns:
            mol: Rdkit sanitized molecule with generated charges and multiple conformers.
        """
        mol = Chem.MolFromSmiles(smiles)
        mol = Chem.AddHs(mol)
        rdForceFieldHelpers.MMFFSanitizeMolecule(mol)
        
        num_rotatable_bonds = Descriptors.NumRotatableBonds(mol)
        amount_of_confs = pow(3, num_rotatable_bonds + 3)
        AllChem.EmbedMultipleConfs(mol, numConfs=amount_of_confs, randomSeed=3407)

        rdPartialCharges.ComputeGasteigerCharges(mol)

        return mol

    @staticmethod
    def set_average_atoms_position(mol,
                                   atoms_idx: list(),
                                   conf_id: int = -1):
        """
        Create virtual atom in the molecule in the middle of atoms with atoms_idx.
        
        Args:
            mol: Rdkit optimized molecule.
            atoms_idx ([int]): Molecules indexes to find their average position.
            conf_id: Conformer id of the molecule. Defaults to -1.
            
        Returns:
            mol: Rdkit molecule with new virtual atom.
            idx (int): index of new virtual atom.
        """
        x, y, z = 0, 0, 0
        for atom_idx in atoms_idx:
            x += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[0]
            y += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[1]
            z += mol.GetConformer(conf_id).GetAtomPosition(atom_idx)[2]
        
        x /= len(atoms_idx)
        y /= len(atoms_idx)
        z /= len(atoms_idx)

        new_atom = rdchem.Atom(0)
        editable_molecule = RWMol(mol)
        idx = editable_molecule.AddAtom(new_atom)
        mol = editable_molecule.GetMol()

        conf = mol.GetConformer(conf_id)
        conf.SetAtomPosition(idx, Point3D(x,y,z))

        return mol, idx

    @staticmethod
    def change_vector_direction(mol, 
                                X1:int, R_1:int, 
                                conf_id:int):
        """
        Changes the direction of the vector X1R-1. Only executed when f_group is "secondary amine":
            1. We have to change direction of X1R-1 to (-1) * X1R1
            2. X1 is in the middle between R(-1) and R1, so R1 = 2*X1 - (R-1)
        
        Args:
            mol: Rdkit optimized molecule.
            X1 (int): Atom index that connects functional group(NH2 or COOH) to the cycle of the molecule.
            R_1 (int): Atom index of the center of the functional group(NH2 or COOH).
            conf_id (int): Conformer id of the molecule. Defaults to -1.
            
        Returns:
            mol: Rdkit molecule with new virtual atom.
            idx (int): index of new virtual atom.
        """
        conf = mol.GetConformer(conf_id)

        Rx_1 = conf.GetAtomPosition(R_1)[0]
        Ry_1 = conf.GetAtomPosition(R_1)[1]
        Rz_1 = conf.GetAtomPosition(R_1)[2]

        Xx1 = conf.GetAtomPosition(X1)[0]
        Xy1 = conf.GetAtomPosition(X1)[1]
        Xz1 = conf.GetAtomPosition(X1)[2]
        
        Rx1 = 2 * Xx1 - Rx_1
        Ry1 = 2 * Xy1 - Ry_1
        Rz1 = 2 * Xz1 - Rz_1

        conf.SetAtomPosition(R_1, Point3D(Rx1,Ry1,Rz1))
        return mol, R_1
    
    @staticmethod
    def is_atom_in_cycle(mol, atom_id):
        """
        Checks if atom is in cycle.
        
        Args:
            mol: Rdkit optimized molecule.
            atom_id (int): Atom id to check if its in cycle.
            
        Returns:
            is_atom_in_ring (bool): True if atom is in ring and False if not.
        """
        atom = mol.GetAtomWithIdx(atom_id)

        is_atom_in_ring = atom.IsInRing()

        return is_atom_in_ring

    def find_X1X2R1R2(self):
        """
        Determines which atoms correspond to X1, X2, R1 and R2
            
        Returns:
            X1 (int): Atom id in cycle, that connects NH2 or COOH to the molecule.
            X2 (int): Atom id in cycle, that connects fluorine functional group to the molecule.
            R1 (int): NH2 or COOH functional group center atom id.
            R2 (int): fluorine functional group center atom id.
        """
        f_group_smiles = self.functional_group_to_smiles[self.f_group]

        carboxile_submol = Chem.MolFromSmiles('CC=O')
        nitro_amine_submol = Chem.MolFromSmiles('CN')

        carboxile_matches = self.mol.GetSubstructMatches(carboxile_submol)
        nitro_amine_matches = self.mol.GetSubstructMatches(nitro_amine_submol)

        X1, R1 = None, None
        if self.identificator == Identificator.carboxilic_acid:
            if len(carboxile_matches) == 0:
                raise "Problem with carboxile acid"
            
            X1 = carboxile_matches[0][0]
            R1 = carboxile_matches[0][1]

        print(self.identificator.name)
        if "amine" in self.identificator.name.lower():
            if len(nitro_amine_matches) == 0:
                raise "Problem with amine"
            
            if "primary" in self.identificator.lower():
                X1 = nitro_amine_matches[0][0]
                R1 = nitro_amine_matches[0][1]
            elif "secondary" in self.identificator.lower():
                X1 = nitro_amine_matches[0][1]
                self.mol, R_1 = OptimizedMolecule.set_average_atoms_position(self.mol, [nitro_amine_matches[0][0], nitro_amine_matches[1][0]], self.min_energy_conf_index)
                self.mol, R1 = OptimizedMolecule.change_vector_direction(self.mol, X1, R_1=R_1, conf_id=self.min_energy_conf_index)

        X2, R2 = None, None
        f_group_submol = Chem.MolFromSmiles(f_group_smiles)
        f_group_matches = self.mol.GetSubstructMatches(f_group_submol)
        if self.f_group.upper() in ['CF3', 'CHF2', 'CH2F']:
            X2 = f_group_matches[0][0]
            R2 = f_group_matches[0][1]

        elif self.f_group == 'gem-CF2':
            X2 = f_group_matches[0][0]
            self.mol, R2 = OptimizedMolecule.set_average_atoms_position(self.mol, [f_group_matches[0][1], f_group_matches[0][2]], self.min_energy_conf_index)

        elif self.f_group.upper() == 'CHF':
            if len(f_group_matches) == 1:
                X2 = f_group_matches[0][0]
                R2 = f_group_matches[0][1]
            elif len(f_group_matches) == 2:
                self.mol, X2 = OptimizedMolecule.set_average_atoms_position(self.mol, [f_group_matches[0][0], f_group_matches[1][0]], self.min_energy_conf_index)
                self.mol, R2 = OptimizedMolecule.set_average_atoms_position(self.mol, [f_group_matches[0][1], f_group_matches[1][1]], self.min_energy_conf_index)

        if len(set([X1, X2, R1, R2])) != 4:
            X1, X2, R1, R2 = None, None, None, None
        
        if not OptimizedMolecule.is_atom_in_cycle(mol=self.mol, atom_id=f_group_matches[0][0]):
            # print(f"X1: {X1} or X2: {X2} is not in the cycle, smiles: {self.smiles}")
            X1, X2, R1, R2 = None, None, None, None

        return X1, X2, R1, R2

    def optimize_geometry(self):
        """Executes geometric optimization of the molecule"""
        self.mol = OptimizedMolecule.prepare_molecule(self.smiles)
        self.min_energy_conf_index, self.min_energy, self.mol = OptimizedMolecule.find_conf_with_min_energy(self.mol)


class DipoleMoment(OptimizedMolecule):
    """
    Class for calculating whole dipole moment of the molecule.

    This class inherits from OptimizedMolecule and provides functionality to calculate dipole moment 
    based on atoms positions and charges using optimized molecule with lowest energy conformer.

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeVolume class.
        calculate(): Calculate the dipole moment of the molecule.
    """
    def __init__(self, smiles):
        super().__init__(smiles)

    def description(self):
        """Returns description of the class"""
        return "Whole dipole moment of the optimized molecule with lowest energy conformer"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the DipoleMoment class."""
        return {"mol": OptimizedMolecule()}
    
    def calculate(self):
        """
        Calculate the dipole moment.

        Returns:
            dipole_moment (float): Float value with calculated dipole_moment
        """
        charges = []
        coordinates = []
        x_centroid, y_centroid, z_centroid = 0, 0, 0
        for atom in self.mol.GetAtoms():
            pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(atom.GetIdx())
            charge = atom.GetDoubleProp("_GasteigerCharge")

            charges.append(charge)
            coordinates.append([pos[0], pos[1], pos[2]])

            x_centroid += pos[0]
            y_centroid += pos[1]
            z_centroid += pos[2]

        x_centroid /= len(self.mol.GetAtoms())
        y_centroid /= len(self.mol.GetAtoms())
        z_centroid /= len(self.mol.GetAtoms())

        charges_multiply_coordinates = coordinates.copy()
        for charges_multiply_coordinate_index in range(len(charges_multiply_coordinates)):
            charges_multiply_coordinates[charges_multiply_coordinate_index][0] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][0] - x_centroid) \
                * charges[charges_multiply_coordinate_index]
            charges_multiply_coordinates[charges_multiply_coordinate_index][1] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][1] - y_centroid) \
                * charges[charges_multiply_coordinate_index]
            charges_multiply_coordinates[charges_multiply_coordinate_index][2] = \
                (charges_multiply_coordinates[charges_multiply_coordinate_index][2] - z_centroid) \
                * charges[charges_multiply_coordinate_index]

        dipole_moment_vector = [0, 0, 0]
        for charges_multiply_coordinate_index in range(len(charges_multiply_coordinates)):
            dipole_moment_vector[0] += charges_multiply_coordinates[charges_multiply_coordinate_index][0]
            dipole_moment_vector[1] += charges_multiply_coordinates[charges_multiply_coordinate_index][1]
            dipole_moment_vector[2] += charges_multiply_coordinates[charges_multiply_coordinate_index][2]

        dipole_moment = math.sqrt(pow(dipole_moment_vector[0], 2) 
                                  + pow(dipole_moment_vector[1], 2)
                                  + pow(dipole_moment_vector[2], 2))

        return dipole_moment


class MoleculeVolume(OptimizedMolecule):
    """
    Class for calculating molecule Volume.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    molecule volume using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeVolume class.
        calculate(): Calculate the molecule volume.
    """
    def __init__(self, smiles):
        super().__init__(smiles)

    def description(self):
        """Returns description of the class"""
        return "Molecule volume on optimized molecule with lowest energy conformer"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the MoleculeVolume class."""
        return {"mol": OptimizedMolecule()}
    
    def calculate(self):
        """
        Calculate the molecule Volume.

        Returns:
            mol_volume (float): Float value with calculated molecule volume
        """
        mol_volume = AllChem.ComputeMolVolume(mol=self.mol,
                                              confId=self.min_energy_conf_index)
        
        return mol_volume


class MoleculeSASA(OptimizedMolecule):
    """
    Class for calculating molecule solvent accessible surface area.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    molecule solvent accessible surface area using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeSASA class.
        calculate(): Calculate the molecule solvent accessible surface area.
    """
    def __init__(self, smiles):
        super().__init__(smiles)

    def description(self):
        """Returns description of the class"""
        return "Molecule free solvent access area on optimized molecule with lowest energy conformer"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the MoleculeSASA class."""
        return {"mol": OptimizedMolecule()}
    
    def calculate(self):
        """
        Calculate the molecule solvent accessible surface area.

        Returns:
            sasa (float): Float value with calculated molecule solvent accessible surface area.
        """
        mol_classify = rdFreeSASA.classifyAtoms(self.mol)
        sasa = rdFreeSASA.CalcSASA(mol=self.mol, 
                                   radii=mol_classify, 
                                   confIdx=self.min_energy_conf_index)
        
        return sasa


class MoleculeTPSAF(OptimizedMolecule):
    """
    Class for calculating molecule topological polar surface area.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    molecule topological polar surface area with additional fluorine area using optimized molecule 
    with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeTPSAF class.
        calculate(): Calculate the molecule topological polar surface area with additional fluorine area.
    """
    def __init__(self, smiles):
        super().__init__(smiles)

    def description(self):
        """Returns description of the class"""
        return "Topological polar surface area with additional Fluor surface area"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the MoleculeTPSAF class."""
        return {"mol": OptimizedMolecule()}
    
    def calculate(self):
        """
        Calculate the molecule topological polar surface area with additional fluorine area.

        Returns:
            tpsa_f (float): Float value with calculated molecule topological polar surface area 
            with additional fluorine area.
        """
        tpsa = Descriptors.TPSA(self.mol)
        fluor_idxs = [atom.GetIdx() for atom in self.mol.GetAtoms() if atom.GetSymbol().lower() == 'f']
        tpsa_f = tpsa

        radii = rdFreeSASA.classifyAtoms(self.mol)
        rdFreeSASA.CalcSASA(self.mol, radii)
        
        for fluor_idx in fluor_idxs:
            atom_sasa = self.mol.GetAtoms()[fluor_idx].GetProp('SASA')
            tpsa_f += float(atom_sasa)

        return tpsa_f


class MoleculeDihedralAngle(OptimizedMolecule):
    """
    Class for calculating dihedral angle between X1, R1, R2 and X2.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    dihedral angle using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeDihedralAngle class.
        calculate(): Calculate the dihedral angle.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "dihedral_angle"

    def description(self):
        """Returns description of the class"""
        return "Molecule free solvent access area on optimized molecule with lowest energy conformer"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the MoleculeDihedralAngle class."""
        return {"mol": OptimizedMolecule()}
    
    @staticmethod
    def _dihedral_angle(mol, 
                       iAtomId:int, jAtomId:int, kAtomId:int, lAtomId:int, 
                       conf_id:int):
        """
        Calculate dihedral angle in molecule between IJK and JKL flat areas with specified conformer id.
        
        Args:
            mol: Rdkit optimized molecule.
            iAtomId (int): AtomI index in molecule.
            jAtomId (int): AtomJ index in molecule.
            kAtomId (int): AtomK index in molecule.
            lAtomId (int): AtomL index in molecule.
            conf_id: Conformer id of the molecule.
            
        Returns:
            dihedral_angle_value (float): dihedral angle between areas.
        """
        
        conf = mol.GetConformer(conf_id)

        dihedral_angle_value = abs(rdMolTransforms.GetDihedralDeg(conf, iAtomId, jAtomId, kAtomId, lAtomId))

        return dihedral_angle_value
    
    def calculate(self):
        """
        Calculate dihedral angle in molecule between IJK and JKL flat areas.

        Returns:
            dihedral_angle_value (float): dihedral angle between areas R2X2X1 and X2X1R1.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate dihedral angle for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")

        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        dihedral_angle_value = MoleculeDihedralAngle._dihedral_angle(self.mol, 
                                                            self.R2, self.X2, self.X1, self.R1, 
                                                            self.min_energy_conf_index) 
        return dihedral_angle_value


class MoleculeDistance(OptimizedMolecule):
    """
    A class for calculating distances between atoms in a molecule.

    This class extends the OptimizedMolecule class to provide methods for calculating
    distances between atoms in a molecule.

    Attributes:
        Inherits attributes from OptimizedMolecule.

    Methods:
        _distance_between_atoms(): Calculates the distance between two atoms.
    """
    def __init__(self, smiles, f_group, identificator):
        """
        Initialize the MoleculeDistance object.

        Args:
            smiles (str): The SMILES string representing the molecule.
            f_group (str): The functional group of the molecule.
            identificator (str): The identifier of the molecule.
        """
        super().__init__(smiles, f_group, identificator)
    
    @staticmethod
    def _distance_between_atoms(iAtom_pos, jAtom_pos):
        """
        Calculate distance between two atoms.
        
        Args:
            iAtom_pos (tuple(int, int, int)): AtomI coordinates.
            jAtom_pos (tuple(int, int, int)): AtomJ coordinates.
            
        Returns:
            distance (float): distance between two points.
        """
        vector = (jAtom_pos[0] - iAtom_pos[0], jAtom_pos[1] - iAtom_pos[1], jAtom_pos[2] - iAtom_pos[2])

        distance = math.sqrt(pow(vector[0], 2) + pow(vector[1], 2) + pow(vector[2], 2))

        return distance


class DistanceBetweenX1X2(MoleculeDistance):
    """
    Class for calculating distance between X1 and X2 atoms in cycles.

    This class inherits from MoleculeDistance and provides functionality to calculate 
    distance using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the DistanceBetweenX1X2 class.
        calculate(): Calculate the distance between X1 and X2.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "distance_between_atoms_in_cycle_and_f_group"

    def description(self):
        """Returns description of the class"""
        return "Distance between atoms, which connect functional groups to the main part of the molecule"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the DistanceBetweenX1X2 class."""
        return {"mol": OptimizedMolecule(), "_distance_between_atoms": MoleculeDistance()}

    def calculate(self):
        """
        Calculate distance in molecule between X1 and X2.

        Returns:
            r_distance (float): dihedral between X1 and X2 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate distance for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        X1_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.X1)
        X2_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.X2)

        r_distance = DistanceBetweenX1X2._distance_between_atoms(X1_pos, X2_pos)

        return r_distance


class DistanceBetweenR1R2(MoleculeDistance):
    """
    Class for calculating distance between R1 and R2 atoms in functional groups centers.

    This class inherits from MoleculeDistance and provides functionality to calculate 
    distance using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the DistanceBetweenR1R2 class.
        calculate(): Calculate the distance between R1 and R2.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "distance_between_atoms_in_f_group_centers"

    def description(self):
        """Returns description of the class"""
        return "Distance between atoms in the centers of functional groups"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the DistanceBetweenR1R2 class."""
        return {"mol": OptimizedMolecule(), "_distance_between_atoms": MoleculeDistance()}

    def calculate(self):
        """
        Calculate distance in molecule between R1 and R2.

        Returns:
            R_distance (float): distance between R1 and R2 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate distance for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        R1_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.R1)
        R2_pos = self.mol.GetConformer(self.min_energy_conf_index).GetAtomPosition(self.R2)

        R_distance = DistanceBetweenR1R2._distance_between_atoms(R1_pos, R2_pos)

        return R_distance


class AngleX1X2R2(OptimizedMolecule):
    """
    Class for calculating flat angle between X1, X2 and R2 atoms.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    flat angle using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the AngleX1X2R2 class.
        _flat_angle(): Calculate flat angle between 3 atoms.
        calculate(): Calculate the angle between X1, X2 and R2 points.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "angle_X1X2R2"

    def description(self):
        """Returns description of the class"""
        return "Flat angle between X1, X2 and R2"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the AngleX1X2R2 class."""
        return {"mol": OptimizedMolecule()}
    
    @staticmethod
    def _flat_angle(mol, 
                    iAtomId:int, jAtomId:int, kAtomId:int, 
                    conf_id:int):
        """
        Calculate flat angle between I, J and K atoms in molecule.
        
        Args:
            iAtomId (int): AtomI coordinates.
            jAtomId (int): AtomJ coordinates.
            kAtomId (int): AtomK coordinates.
            
        Returns:
            (float): IJK Angle.
        """
        conf = mol.GetConformer(conf_id)

        return rdMolTransforms.GetAngleDeg(conf, iAtomId, jAtomId, kAtomId)


    def calculate(self):
        """
        Calculate angle in molecule between X1, X2 and R2.

        Returns:
            angle_X1X2R2 (float): angle between X1, X2 and R2 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate angle for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        angle_X1X2R2 = AngleX1X2R2._flat_angle(self.mol, self.X1, self.X2, self.R2, self.min_energy_conf_index)

        return angle_X1X2R2


class AngleX2X1R1(OptimizedMolecule):
    """
    Class for calculating flat angle between X2, X1 and R1 atoms.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    flat angle using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the AngleX2X1R1 class.
        _flat_angle(): Calculate flat angle between 3 atoms.
        calculate(): Calculate the angle between X2, X1 and R1 points.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "angle_X2X1R1"

    def description(self):
        """Returns description of the class"""
        return "Flat angle between X2, X1 and R1"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the AngleX2X1R1 class."""
        return {"mol": OptimizedMolecule()}
    
    @staticmethod
    def _flat_angle(mol, 
                    iAtomId:int, jAtomId:int, kAtomId:int, 
                    conf_id:int):
        """
        Calculate flat angle between I, J and K atoms in molecule.
        
        Args:
            iAtomId (int): AtomI coordinates.
            jAtomId (int): AtomJ coordinates.
            kAtomId (int): AtomK coordinates.
            
        Returns:
            (float): IJK Angle.
        """
        conf = mol.GetConformer(conf_id)

        return rdMolTransforms.GetAngleDeg(conf, iAtomId, jAtomId, kAtomId)


    def calculate(self):
        """
        Calculate angle in molecule between X2, X1 and R1.

        Returns:
            angle_X2X1R1 (float): angle between X2, X1 and R1 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate angle for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        angle_X2X1R1 = AngleX2X1R1._flat_angle(self.mol, self.X2, self.X1, self.R1, self.min_energy_conf_index)

        return angle_X2X1R1


class AngleR2X2R1(OptimizedMolecule):
    """
    Class for calculating flat angle between R2, X2 and R1 atoms.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    flat angle using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the AngleR2X2R1 class.
        _flat_angle(): Calculate flat angle between 3 atoms.
        calculate(): Calculate the angle between R2, X2 and R1 points.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "angle_R2X2R1"

    def description(self):
        """Returns description of the class"""
        return "Flat angle between R2, X2 and R1"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the AngleR2X2R1 class."""
        return {"mol": OptimizedMolecule()}
    
    @staticmethod
    def _flat_angle(mol, 
                    iAtomId:int, jAtomId:int, kAtomId:int, 
                    conf_id:int):
        """
        Calculate flat angle between I, J and K atoms in molecule.
        
        Args:
            iAtomId (int): AtomI coordinates.
            jAtomId (int): AtomJ coordinates.
            kAtomId (int): AtomK coordinates.
            
        Returns:
            (float): IJK Angle.
        """
        conf = mol.GetConformer(conf_id)

        return rdMolTransforms.GetAngleDeg(conf, iAtomId, jAtomId, kAtomId)


    def calculate(self):
        """
        Calculate angle in molecule between R2, X2 and R1.

        Returns:
            angle_X2X1R1 (float): angle between R2, X2 and R1 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate angle for this molecule.
        """
        if self.f_group is None or self.identificator is None:
            raise ValueError("fluor functional group or molecule identificator NotFound")
        
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        angle_R2X2R1 = AngleR2X2R1._flat_angle(self.mol, self.R2, self.X2, self.R1, self.min_energy_conf_index)

        return angle_R2X2R1


class AngleR1X1R2(OptimizedMolecule):
    """
    Class for calculating flat angle between R1, X1 and R2 atoms.

    This class inherits from OptimizedMolecule and provides functionality to calculate 
    flat angle using optimized molecule with lowest energy conformer and rdkit.

    Attributes:
        smiles (str): String representation of a molecule.
        f_group (str): Fluor functional group name. Defaults to None.
        identificator (Identificator): The molecule type. Defaults to None.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the AngleR1X1R2 class.
        _flat_angle(): Calculate flat angle between 3 atoms.
        calculate(): Calculate the angle between R1, X1 and R2 points.
    """
    def __init__(self, smiles, f_group, identificator):
        super().__init__(smiles, f_group, identificator)
        self.feature_name = "angle_R1X1R2"

    def description(self):
        """Returns description of the class"""
        return "Flat angle between R1, X1 and R2"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the AngleR1X1R2 class."""
        return {"mol": OptimizedMolecule()}
    
    @staticmethod
    def _flat_angle(mol, 
                    iAtomId:int, jAtomId:int, kAtomId:int, 
                    conf_id:int):
        """
        Calculate flat angle between I, J and K atoms in molecule.
        
        Args:
            iAtomId (int): AtomI coordinates.
            jAtomId (int): AtomJ coordinates.
            kAtomId (int): AtomK coordinates.
            
        Returns:
            (float): IJK Angle.
        """
        conf = mol.GetConformer(conf_id)

        return rdMolTransforms.GetAngleDeg(conf, iAtomId, jAtomId, kAtomId)


    def calculate(self):
        """
        Calculate angle in molecule between R1, X1 and R2.

        Returns:
            angle_R1X1R2 (float): angle between R1, X1 and R2 atoms.
        
        Raises:
            ValueError: If the f_group or identificator is None.
            InvalidMoleculeTypeError: If we cannot calculate angle for this molecule.
        """
        if len(set([self.X1, self.X2, self.R1, self.R2])) != 4:
            raise InvalidMoleculeTypeError(self.X1, self.X2, self.R1, self.R2, self.feature_name)

        angle_R1X1R2 = AngleR2X2R1._flat_angle(self.mol, self.R1, self.X1, self.R2, self.min_energy_conf_index)

        return angle_R1X1R2


class Molecule2D:
    """
    Class that represents a 3D molecule with single conformer.

    This class provides functionality to convert smiles to rdkit molecule, 

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        _prepare_molecule(): Creates 3D molecule with single conformers from smiles.
    """
    def __init__(self,
                 smiles) -> None:
        
        self.smiles = smiles

        self.mol = Molecule2D._prepare_molecule(SMILES)
    
    @staticmethod
    def _prepare_molecule(SMILES):
        """
        Create rdkit 3d molecule from SMILES with conformer.
        
        Args:
            SMILES (str): String representation of a molecule. 
            
        Returns:
            mol: Rdkit 3D molecule with conformer.
        """
        mol = Chem.MolFromSmiles(SMILES)
        mol = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol, randomSeed=42)

        return mol


class MoleculeRingsAmount(Molecule2D):
    """
    Class for calculating amount of rings in the molecule.

    This class inherits from Molecule2D and provides functionality 
    to calculate amount of rings in the molecule 

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeVolume class.
        _cycles_amount(): Get amount of rings in the molecule.
        calculate(): Get ring number in the molecule.
    """
    def __init__(self, smiles):
        super().__init__(smiles)
        self.feature_name = "mol_num_cycles"

    def description(self):
        """Returns description of the class"""
        return "Molecule's amount of cycles"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the MoleculeRingsAmount class."""
        return {"mol": Molecule2D(), "num_rings": cycles_amount()}

    @staticmethod
    def _cycles_amount(mol):
        """
        Calculate amount of cycles in the molecule.
        
        Args:
            mol: rdkit molecule.
            
        Returns:
            num_rings (int): amount of rings.
        """
        sssr = Chem.GetSSSR(mol)

        num_rings = len(sssr)
        
        return num_rings
    
    def calculate(self):
        """
        Calculate amount of cycles in the molecule.
            
        Returns:
            num_rings (int): amount of rings.
        """
        num_rings = MoleculeRingsAmount._cycles_amount(self.mol)
        
        return num_rings


class AtomsToRingRatio(Molecule2D):
    """
    Class for calculating ratio of amount of atoms in rings to rings amount.

    This class inherits from Molecule2D and provides functionality 
    to calculate amount of atoms in rings / amount of rings. 

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeVolume class.
        _cycles_amount(): Get amount of rings in the molecule.
        _atoms_in_cycles_amount(): Get amount of atoms in cycles in the molecule.
        calculate(): Get ring number in the molecule.
    """
    def __init__(self, smiles):
        super().__init__(smiles)
        self.feature_name = "avg_atoms_in_cycle"

    def description(self):
        """Returns description of the class"""
        return "Molecule's amount of atoms in cycle divide by amount of cycles"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the AtomsToRingRatio class."""
        return {"mol": Molecule2D()}
    
    @staticmethod
    def _cycles_amount(mol):
        """
        Calculate amount of cycles in the molecule.
        
        Args:
            mol: rdkit molecule.
            
        Returns:
            amount_of_cycles (int): amount of rings.
        """
        sssr = Chem.GetSSSR(mol)
        
        amount_of_cycles = len(sssr)
        
        return amount_of_cycles
    
    @staticmethod
    def _atoms_in_cycles_amount(mol):
        """
        Calculate amount of atoms that are in cycle in molecule.
        
        Args:
            mol: rdkit molecule.
            
        Returns:
            atoms_num_in_cycles (int): amount of atoms in cycle in the molecule.
        """
        sssr = Chem.GetSSSR(mol)
        
        atoms_idxs = set()

        for i, ring in enumerate(sssr):
            for atom_index in ring:
                atoms_idxs.add(atom_index)
        
        atoms_num_in_cycles = len(atoms_idxs)

        return atoms_num_in_cycles
    
    def calculate(self):
        """
        Calculate ratio amount of atoms in rings to rings amount.
            
        Returns:
            (float): atoms_num_in_cycles / amount_cycles.
        """
        amount_cycles = AtomsToRingRatio._cycles_amount(self.mol)
        if amount_cycles == 0:
            return 0
        
        atoms_num_in_cycles = AtomsToRingRatio._atoms_in_cycles_amount(self.mol)
        
        return atoms_num_in_cycles / amount_cycles


class Chirality(Molecule2D):
    """
    Class for calculating amount of chiral centers in molecule.

    This class inherits from Molecule2D and provides functionality 
    to calculate amount of chiral centers. 

    Attributes:
        smiles (str): String representation of a molecule.

    Methods:
        description(): Description of the class.
        dependencies(): Dependencies needed for the MoleculeVolume class.
        _amount_of_chiral_centers(): Get amount of chiral centers in the molecule.
        calculate(): Get chirals centers amount..
    """
    def __init__(self, smiles):
        super().__init__(smiles)
        self.feature_name = "chirality"

    def description(self):
        """Returns description of the class"""
        return "Amount of chiral centers in the molecule"
    
    def dependencies(self):
        """Returns a list of dependencies needed for the Chirality class."""
        return {"mol": Molecule2D()}
    
    @staticmethod
    def _amount_of_chiral_centers(mol):
        """
        Calculate amount of chiral centers in molecule.
        
        Args:
            mol: rdkit molecule.
            
        Returns:
            amount_of_chiral_centers (int): amount of chiral centers in the molecule.
        """
        Chem.AssignAtomChiralTagsFromStructure(mol)
        
        chirality_centers = Chem.FindMolChiralCenters(mol)

        amount_of_chiral_centers = len(chirality_centers)
        
        return amount_of_chiral_centers
    
    def calculate(self):
        """
        Calculate amount of chiral centers in molecule.
            
        Returns:
            amount_of_chiral_centers (int): amount of chiral centers in the molecule.
        """
        amount_of_chiral_centers = Chirality._amount_of_chiral_centers(self.mol)
        
        return amount_of_chiral_centers


if __name__ == "__main__":
    SMILES = "FC1(F)CCC(C(O)=O)CC1"
    identificator = Identificator.carboxilic_acid
    f_group = "gem-CF2"
    feature = DistanceBetweenX1X2(smiles=SMILES,
                                  f_group=f_group,
                                  identificator=identificator)
    print(feature.calculate())



from rdkit import Chem
from mordred import Calculator, descriptors
from rdkit.Chem import Descriptors, AllChem, rdForceFieldHelpers, rdPartialCharges

from fluoriclogppka.ml_part.services.utils import has_numbers

class MordredFeaturesService:
    """
    A service class for obtaining Mordred features from chemical structures.

    This class provides methods to prepare a molecule from a SMILES string and
    obtain Mordred features for the molecule.

    Attributes:
        mol: The molecule object prepared from the SMILES string.
        mordred_features_dict: A dictionary containing Mordred features extracted
            from the molecule.

    Methods:
        __init__(): Initializes the MordredFeaturesService object.
        prepare_molecule(): Prepares a molecule object from a SMILES string.
        obtain_mordred_features(): Obtains Mordred features for the molecule.
    """
    def __init__(self,
                 SMILES):
        """
        Initialize the MordredFeaturesService object.

        Args:
            SMILES (str): The SMILES string representing the molecule.
        """
        self.mol = MordredFeaturesService.prepare_molecule(SMILES)

        self.mordred_features_dict = self.obtain_mordred_features()


    @staticmethod
    def prepare_molecule(SMILES):
        """
        Prepare a molecule object from a SMILES string.

        Args:
            SMILES (str): The SMILES string representing the molecule.

        Returns:
            Mol: The prepared molecule object.
        """
        mol = Chem.MolFromSmiles(SMILES)
        mol = Chem.AddHs(mol)

        AllChem.EmbedMolecule(mol, randomSeed=42)

        return mol


    def obtain_mordred_features(self):
        """
        Obtain Mordred features for the molecule.

        Returns:
            dict: A dictionary containing Mordred features extracted from the molecule.
        """
        calc = Calculator(descriptors, ignore_3D=False)
        df = calc.pandas([self.mol], quiet=True)

        mordred_dict = {}
        for _, row in df.iterrows():
            for key in row.keys():
                if "ring" in key.lower() and has_numbers(key):
                    continue
                if type(row[key]) in [int, float]:
                    mordred_dict[key] = row[key]

        return mordred_dict



import pandas as pd

import warnings
import h2o
from h2o.exceptions import H2ODependencyWarning

warnings.filterwarnings("ignore", category=H2ODependencyWarning)

from fluoriclogppka.ml_part.constants import Target
from fluoriclogppka.ml_part.constants import LOGP_MODEL_PATH, PKA_AMINE_MODEL_PATH, PKA_ACID_MODEL_PATH

class H2OService:
    """
    A service class for working with H2O models.

    This class provides methods for initializing an H2O model, converting data in 
    specific h2o format for prediction, and making predictions using the model.

    Attributes:
        model_path (str): The path to the pre-trained H2O model.

    Methods:
        _model_init(): Initializes the H2O model from the specified path.
        _prepare_h2o_data(): Prepares H2O data frame from a dictionary of features.
        predict(): Makes predictions using the loaded model.
    """
    def __init__(self,
                 model_path: str):
        """
        Initialize the H2OService object.

        Args:
            model_path (str): The path to the pre-trained H2O model.
        """
        self.model_path = model_path
        
        self.model = H2OService._model_init(self.model_path)
    
    @staticmethod
    def _model_init(model_path):
        """
        Initialize the H2O model from the specified path.

        Args:
            model_path (str): The path to the pre-trained H2O model.

        Returns:
            h2o_model: The initialized H2O model.
        """
        h2o.init(verbose=False)
        h2o.no_progress()

        h2o_model = h2o.load_model(model_path)

        return h2o_model
    
    @staticmethod
    def _prepare_h2o_data(features_dict: dict):
        """
        Prepare H2O data frame from a dictionary of features.

        Args:
            features_dict (dict): A dictionary containing features.

        Returns:
            h2o_frame: The prepared H2O data frame.
        """
        features_df = pd.DataFrame(features_dict, index=[0])
        
        h2o_frame = h2o.H2OFrame(features_df)

        return h2o_frame
    
    def predict(self,
                features_dict):
        """
        Make predictions using the loaded model.

        Args:
            features_dict (dict): A dictionary containing features.

        Returns:
            predictions (float): Predicted target value using h2o model.
        """
        h2o_frame = H2OService._prepare_h2o_data(features_dict)

        predictions_h2o_frame = self.model.predict(h2o_frame)

        predictions = predictions_h2o_frame.as_data_frame()['predict'][0]

        return predictions
